# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import subprocess

from typing import NamedTuple
from pathlib import Path

base_dir = Path(__file__).parents[1]
pyproject_file = str(base_dir  / "pyproject.toml")
requirements_file = str(base_dir / "requirements.txt")
constraints_file = str(base_dir / "constraints.txt")
override_file = str(base_dir / "overrides.txt")


constraint_pattern = re.compile(
    R"(?P<package>.*?==\d+\.\d+\.?\d*)\s*;\s*(?P<constraints>.*?)\s*\\"
)


class ConstraintIssue(NamedTuple):
    package: str
    constraints: str


def check_version_constraints() -> None:
    # Check the requirements file for python version constraints
    discovered_constraints = []
        
    with open(requirements_file, 'r') as f:
        for line in f:
            if match := constraint_pattern.match(line):
                discovered_constraints.append(
                    ConstraintIssue(match["package"], match["constraints"])
                )                  
    if discovered_constraints:
        constraint_msg = "\n".join(
            f"\t{c.package!r}: {c.constraints!r}" for c in discovered_constraints
        )
        
        raise ValueError(f"Packages discovered with banned constraints:\n{constraint_msg}")


def upgrade_requirements():
    subprocess.run(
        [
            "uv", "pip", "compile",
            pyproject_file,
            "--universal",
            "--generate-hashes",
            "-o", requirements_file,
            # "--constraint", constraints_file,
            # "--override", override_file,
            "--upgrade",
        ],
    )


if __name__ == "__main__":
    upgrade_requirements()
    check_version_constraints()



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os
import sys
from unittest.mock import patch, mock_open

import pytest

from ducktools.pytui import __version__
from ducktools.pytui.shells import _core as core, get_shell_script
from ducktools.pytui.shells.bash import BashShell, GitBashShell
from ducktools.pytui.platform_paths import SHELL_SCRIPT_FOLDER


windows_only = pytest.mark.skipif(sys.platform != "win32", reason="Windows Only")
non_windows = pytest.mark.skipif(sys.platform == "win32", reason="Non-Windows Only")


def test_get_shell_script_output():
    # Test the output value at least is as expected
    bash_script = "activate_pytui.sh"

    # Mock that the version *IS* the latest one
    open_mock = mock_open(read_data=__version__)

    with patch("builtins.open", open_mock) as m, \
            patch("os.path.exists") as exists_mock, \
            patch("shutil.rmtree") as rmtree_mock:

        exists_mock.return_value = True
        expected = os.path.join(SHELL_SCRIPT_FOLDER, bash_script)

        result = get_shell_script(bash_script)

    assert result == expected

    exists_mock.assert_called_once_with(expected)
    m.assert_called_once_with(os.path.join(SHELL_SCRIPT_FOLDER, ".version"))
    rmtree_mock.assert_not_called()


class TestDedupePath:
    @non_windows
    def test_linux_path(self):
        venv_path = "/home/david/src/project/.venv"
        demo_path = (
            "/home/david/src/project/.venv:"
            "/home/david/.local/bin:"
            "/home/david/.local/bin:" # Pretty sure this is UV putting this in twice
            "/usr/local/sbin:"
            "/usr/local/bin:"
            "/usr/sbin:"
            "/usr/bin:"
            "/sbin:"
            "/bin:"
            "/usr/local/bin:"  # Duplicate that should be removed
        )

        shell = core.Shell("/bin/bash")

        deduped_expected = (
            "/home/david/src/project/.venv:"
            "/home/david/.local/bin:"
            "/usr/local/sbin:"
            "/usr/local/bin:"
            "/usr/sbin:"
            "/usr/bin:"
            "/sbin:"
            "/bin" # Trailing colon is gone in deduped
        )

        deduped = shell.get_deduped_path(demo_path, venv_path)

        assert deduped == deduped_expected

    @windows_only
    def test_windows_path(self):
        # I will change these back from R to r if I switch from VSCode
        venv_path = R"C:\Users\David\Source\result\.venv"
        demo_path = (
            R"C:\Windows\system32;"
            R"C:\Users\David\Source\result\.venv;"  # Should be removed
            R"C:\Users\David\.local\bin;"
            R"C:\Users\David\.local\bin\;"  # Duplicate should be removed
            R"C:\Program Files\Github CLI\;"
        )

        shell = core.Shell(R"C:\Program Files\PowerShell\7\pwsh.exe")
        deduped = shell.get_deduped_path(demo_path, venv_path)

        assert deduped == (
            R"C:\Users\David\Source\result\.venv;"
            R"C:\Windows\system32;"
            R"C:\Users\David\.local\bin;"
            R"C:\Program Files\Github CLI"
        )


def test_from_path_missing():
    assert core.Shell.from_path("unknown/path") is None


class TestGetDefault:
    @windows_only
    def test_get_default_win_envvar(self):
        # Test with an environment variable
        bash_path = R"C:\Program Files\Git\usr\bin\bash.exe"
        fake_env = {"SHELL": bash_path}
        with patch.dict("os.environ", fake_env, clear=True):
            shell = core.Shell.get_default()
            assert shell == GitBashShell(bash_path)

    @pytest.mark.parametrize("shell", core.Shell.registry.values())
    def test_get_default_which(self, shell):
        # Test with no environment variable
        def fake_which(path):
            if path == shell.bin_name:
                return path
            return None

        with patch.dict("os.environ", {}, clear=True), \
                patch("shutil.which") as mock_which:
            mock_which.side_effect = fake_which

            detected_shell = core.Shell.get_default()
            assert detected_shell == shell(shell.bin_name)

    @windows_only
    def test_backup_comspec(self):
        from ducktools.pytui.shells.cmd import CMDShell
        # Windows will use os.environ["COMSPEC"] if no shell is found on PATH
        cmd_path = R"C:\Windows\system32\cmd.exe"
        fake_env = {"COMSPEC": cmd_path}
        with patch.dict("os.environ", fake_env, clear=True), \
                patch("shutil.which") as mock_which:
            mock_which.return_value = None

            detected_shell = core.Shell.get_default()
            assert detected_shell == CMDShell(cmd_path)

    @non_windows
    def test_get_default_linux_envvar(self):
        bash_path = "/bin/bash"
        fake_env = {"SHELL": bash_path}
        with patch.dict("os.environ", fake_env, clear=True):
            shell = core.Shell.get_default()
            assert shell == BashShell(bash_path)



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os, os.path
import sys

from ducktools.lazyimporter import LazyImporter, FromImport

from . import _lazy_imports as _laz  # Shared LazyImporter for external imports


_laz_internal = LazyImporter(
    [
        FromImport("._version", "__version__", "app_version"),
        FromImport(".config", "Config"),
        FromImport(".ui", "ManagerApp"),
    ],
    globs=globals(),
)


class UnsupportedPythonError(Exception):
    pass


def _check_windows_dir() -> None:
    # Double-clicking to launch the zipapp may put the user in the system32
    # folder on Windows.
    # Inside a zipapp, __file__ won't exist on the file system as it is inside the archive.
    if not os.path.exists(__file__) and (win_path := os.environ.get("WINDIR")):
        sys32 = os.path.join(win_path, "System32")
        if os.path.samefile(os.getcwd(), sys32) and (user_path := os.environ.get("USERPROFILE")):
            os.chdir(user_path)


def get_parser() -> _laz.argparse.ArgumentParser:

    parser = _laz.argparse.ArgumentParser(
        prog="ducktools-pytui",
        description="Prototype Python venv and runtime manager",
    )
    parser.add_argument("-V", "--version", action="version", version=_laz_internal.app_version)

    # Config parser arguments
    subparsers = parser.add_subparsers(title="Subcommands", dest="subcommand")
    config_parser = subparsers.add_parser("config", help="Subcommand for setting config options")

    config_parser.add_argument(
        "--set-shell",
        action="store",
        metavar="SHELL_NAME or SHELL_PATH",
        help="Set the shell to be used for launching activated environments"
    )

    config_parser.add_argument(
        "--set-search-mode",
        action="store",
        choices=_laz_internal.Config.VENV_SEARCH_MODES,
        help="Set the search mode to be used",
    )

    config_parser.add_argument(
        "--set-global-venv-dir",
        action="store",
        help="Set the global venv folder",
    )

    include_pip_group = config_parser.add_mutually_exclusive_group()
    include_pip_group.add_argument(
        "--include-pip",
        dest="include_pip",
        action="store_true",
        default=None,
        help="Include pip when creating venv"
    )
    include_pip_group.add_argument(
        "--exclude-pip",
        dest="include_pip",
        action="store_false",
        default=None,
        help="Exclude pip when creating a venv"
    )

    update_pip_group = config_parser.add_mutually_exclusive_group()
    update_pip_group.add_argument(
        "--update-pip",
        dest="update_pip",
        action="store_true",
        default=None,
        help="Update pip to the latest version when creating a venv"
    )
    update_pip_group.add_argument(
        "--bundled-pip",
        dest="update_pip",
        action="store_false",
        default=None,
        help="Use the version of pip bundled with the runtime when creating a venv"
    )

    return parser


def main() -> int:
    if not sys.stdout.isatty():
        raise RuntimeError("No TTY detected, exiting")

    if sys.version_info < (3, 10):
        v = sys.version_info
        print(
            f"Unsupported Python: {v.major}.{v.minor}.{v.micro} is not supported. "
            f"PyTUI requires Python 3.10 or later."
        )
        input("Press ENTER to close")
        sys.exit()

    if sys.argv[1:]:
        parser = get_parser()
        args = parser.parse_args()

        if args.subcommand == "config":
            from .config import Config
            config = Config.from_file()

            update_config = False
            if (shell := args.set_shell) is not None:
                new_shell = config.set_shell(shell)
                if new_shell:
                    print(f"Default shell set to '{new_shell}'")
                    update_config = True
                else:
                    print(f"Shell '{shell}' could not be found.")

            if (search_mode := args.set_search_mode) is not None:
                update_config = True
                config.venv_search_mode = search_mode
                print(f"venv search mode set to '{search_mode}'")

            if (venv_path := args.set_global_venv_dir) is not None:
                update_config = True
                venv_path = os.path.expanduser(venv_path)
                config.global_venv_folder = venv_path
                print(f"Global venv folder set to \"{venv_path}\"")

            if (include_pip := args.include_pip) is not None:
                update_config = True
                config.include_pip = args.include_pip
                if include_pip:
                    print("New venvs will be created WITH pip")
                else:
                    print("New venvs will be created WITHOUT pip")

            if (update_pip := args.update_pip) is not None:
                update_config = True
                config.latest_pip = update_pip
                if update_pip:
                    print("New venvs with pip will update to the latest pip")
                else:
                    print("New venvs with pip will use the bundled pip")

            if update_config:
                config.write_config()
            else:
                import json
                from ducktools.classbuilder.prefab import as_dict

                print(f"Config file: \"{config.config_file}\"")
                print("Current Settings:")

                data = json.dumps(as_dict(config), indent=2)
                print(data)
                print("\nFor editing options, check '--help'")

    else:
        if sys.platform == "win32":
            _check_windows_dir()

        app = _laz_internal.ManagerApp()
        app.run()

    return 0


if __name__ == "__main__":
    sys.exit(main())



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os
import os.path
import sys

import asyncio
import functools
import subprocess
import sysconfig

from typing import overload, TYPE_CHECKING

from ducktools.pythonfinder import PythonInstall
from ducktools.pythonfinder.venv import list_python_venvs, PythonVEnv, PythonPackage

from textual import work, markup
from textual.app import App
from textual.binding import Binding
from textual.containers import Vertical
from textual.markup import escape
from textual.screen import ModalScreen
from textual.validation import Length
from textual.widgets import Button, DataTable, Footer, Header, Input, Label
from textual.widgets.data_table import CellDoesNotExist


from ._version import __version__ as app_version
from .commands import launch_repl, launch_shell, create_venv, delete_venv
from .config import Config
from .util import list_installs_deduped
from .runtime_installers import (
    PythonListing,
    fetch_downloads,
    find_matching_listing,
    get_managers,
)


CWD = os.getcwd()
HOME = os.environ.get("HOME")

# This mapping handles nicer user facing names for "managed by"
MANAGED_BY_MAPPING = {
    "Astral": "Astral uv",  # UV Python Installer
    "PythonCore": "python.org",  # Windows installer
}

if TYPE_CHECKING:
    @overload
    def substitute_home(p: None, homedir: str | None = ...) -> None: ...

    @overload
    def substitute_home(p: str, homedir: str | None = ...) -> str: ...


def substitute_home(p, homedir=HOME):
    if (
        p
        and sys.platform != "win32"
        and homedir
        and os.path.commonpath([p, homedir]) == homedir
    ):
        relpath = os.path.relpath(p, start=homedir)
        p = os.path.join("~", relpath)
    return p


class InstallableRuntimeTable(DataTable):
    def __init__(self, runtimes: list[PythonListing], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.runtimes = runtimes

    def on_mount(self):
        self.setup_columns()
        self.list_downloads()

    def setup_columns(self):
        self.cursor_type = "row"
        self.add_columns("Version", "Manager", "Implementation", "Variant", "Architecture")

    def list_downloads(self):
        for dl in self.runtimes:
            manager = MANAGED_BY_MAPPING.get(
                dl.manager.organisation,
                dl.manager.organisation
            )

            self.add_row(
                dl.version,
                manager,
                dl.implementation,
                dl.variant,
                dl.arch,
                key=dl.full_key,
            )


class RuntimeInstallScreen(ModalScreen[PythonListing | None]):
    BINDINGS = [
        Binding(key="enter", action="install", description="Install Runtime", priority=True, show=True),
        Binding(key="escape", action="cancel", description="Cancel", show=True),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.runtimes = fetch_downloads()
        self.install_table = InstallableRuntimeTable(self.runtimes, classes="boxed_noborder")

        self.install_button = Button("Install", variant="success", id="install")
        self.cancel_button = Button("Cancel", id="cancel")

    @property
    def runtimes_by_key(self):
        return {
            v.full_key: v for v in self.runtimes
        }

    @property
    def selected_runtime(self) -> PythonListing | None:
        table = self.install_table

        try:
            row = table.coordinate_to_cell_key(table.cursor_coordinate)
        except CellDoesNotExist:
            return None

        return self.runtimes_by_key.get(row.row_key.value)

    def compose(self):
        with Vertical(classes="boxed"):
            yield Label("Available Python Runtimes")
            yield self.install_table
            yield Footer()

    def action_install(self):
        if self.focused == self.install_table or self.focused == self.install_button:
            self.dismiss(self.selected_runtime)
        else:
            self.dismiss(None)

    def action_cancel(self):
        self.dismiss(None)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "install":
            self.dismiss(self.selected_runtime)
        else:
            self.dismiss(None)


class DependencyScreen(ModalScreen[list[PythonPackage]]):
    BINDINGS = [
        Binding(key="r", action="reload_dependencies", description="Reload Dependencies", show=True),
        Binding(key="escape", action="close", description="Close", show=True),
    ]

    venv: PythonVEnv
    dependency_cache: list[PythonPackage] | None
    venv_table: DataTable

    def __init__(
        self,
        venv: PythonVEnv,
        dependency_cache: list[PythonPackage] | None
    ):
        super().__init__()
        self.venv = venv
        self.dependency_cache = dependency_cache

        self.venv_table = DataTable(classes="boxed_noborder")

    def compose(self):
        with Vertical(classes="boxed"):
            yield Label(f"Packages installed in {self.venv.folder}")
            yield self.venv_table
            yield Footer()

    def on_mount(self):
        self.venv_table.cursor_type = "row"
        self.venv_table.add_columns("Dependency", "Version")
        self.load_dependencies()

    def action_close(self):
        self.dismiss(self.dependency_cache)

    async def action_reload_dependencies(self):
        self.load_dependencies(clear=True)

    @work
    async def load_dependencies(self, clear=False):
        self.venv_table.loading = True
        try:
            if clear:
                self.venv_table.clear()
                self.dependency_cache = None

            dependencies = self.dependency_cache
            if dependencies is None:
                loop = asyncio.get_running_loop()
                dependencies = await loop.run_in_executor(None, self.venv.list_packages)

            for dep in sorted(dependencies, key=lambda x: x.name.lower()):
                self.venv_table.add_row(dep.name, dep.version, key=dep.name)
        finally:
            self.venv_table.loading = False

        self.dependency_cache = dependencies


class VEnvCreateScreen(ModalScreen[str | None]):
    BINDINGS = [
        Binding(key="enter", action="create", description="Create VEnv", show=True, priority=True),
        Binding(key="escape", action="cancel", description="Cancel", show=True),
    ]

    def __init__(self, runtime: PythonInstall, global_venv=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.runtime = runtime
        self.global_venv = global_venv

        if global_venv:
            self.venv_input = Input(
                placeholder="VEnv Path",
                validators=[Length(minimum=1)],
                validate_on=["submitted"],
            )
        else:
            self.venv_input = Input(placeholder="VEnv Path (default='.venv')")


    def compose(self):
        with Vertical(classes="boxed"):
            yield Label(f"Create VENV from {self.runtime.implementation} {self.runtime.version_str}")
            with Vertical(classes="boxed_noborder"):
                yield self.venv_input
            yield Footer()

    def action_cancel(self):
        self.dismiss(None)

    def action_create(self):
        if self.global_venv and len(self.venv_input.value) == 0:
            self.notify("Must provide a name in order to create a global venv", severity="warning")
        else:
            self.dismiss(self.venv_input.value)

    def on_input_submitted(self, event: Input.Submitted):
        self.dismiss(event.value)


class VEnvTable(DataTable):
    BINDINGS = [
        Binding(key="enter", action="app.activated_shell", description="Launch VEnv Shell", show=True),
        Binding(key="r", action="app.launch_venv_repl", description="Launch VEnv REPL", show=True),
        Binding(key="p", action="app.list_venv_packages", description="List Packages", show=True),
        Binding(key="delete", action="app.delete_venv", description="Delete VEnv", show=True),
    ]

    def __init__(self, *args, config, **kwargs):
        super().__init__(*args, **kwargs)

        self.config = config

        self._venv_catalogue = {}

    def on_mount(self):
        self.setup_columns()
        self.call_after_refresh(self.load_venvs, clear_first=False)

    def setup_columns(self):
        self.cursor_type = "row"
        self.add_columns("Version", "Global", "Environment Path", "Runtime Path")

    @staticmethod
    def _keysort(rowtuple):
        return rowtuple[1], rowtuple[2]

    def sort_by_path(self):
        self.sort(key=self._keysort)

    def venv_from_key(self, key) -> PythonVEnv:
        return self._venv_catalogue[key]

    def add_venv(self, venv: PythonVEnv, sort=False, global_venv=False):
        self._venv_catalogue[venv.folder] = venv

        if global_venv:
            self.add_row(
                venv.version_str,
                True,
                substitute_home(venv.folder),
                substitute_home(venv.parent_executable),
                key=venv.folder
            )
        else:
            folder = os.path.relpath(venv.folder, start=CWD)
            self.add_row(
                venv.version_str,
                False,
                folder,
                substitute_home(venv.parent_executable),
                key=venv.folder
            )
        if sort:
            self.sort_by_path()

    def remove_venv(self, venv: PythonVEnv):
        self.remove_row(row_key=venv.folder)
        self._venv_catalogue.pop(venv.folder)

    @work
    async def load_venvs(self, clear_first=True):
        self.loading = True
        try:
            if clear_first:
                self.clear(columns=False)
                self._venv_catalogue = {}

            recursive = "recursive" in self.config.venv_search_mode
            search_parent_folders = "parents" in self.config.venv_search_mode

            global_venv_folder = self.config.global_venv_folder

            loop = asyncio.get_running_loop()
            get_venvs = functools.partial(
                list_python_venvs,
                base_dir=CWD,
                recursive=recursive,
                search_parent_folders=search_parent_folders
            )

            venvs = await loop.run_in_executor(
                None,
                get_venvs,
            )

            for venv in venvs:
                if not os.path.commonpath([venv.folder, global_venv_folder]) == global_venv_folder:
                    self.add_venv(venv, sort=False)

            if os.path.exists(self.config.global_venv_folder):
                get_global_venvs = functools.partial(
                    list_python_venvs,
                    base_dir=self.config.global_venv_folder,
                    recursive=True,
                    search_parent_folders=False,
                )

                global_venvs = await loop.run_in_executor(
                    None,
                    get_global_venvs
                )

                for venv in global_venvs:
                    self.add_venv(venv, global_venv=True, sort=False)

        finally:
            self.sort_by_path()
            self.refresh_bindings()
            self.loading = False


class RuntimeTable(DataTable):
    BINDINGS = [
        Binding(key="r", action="app.launch_runtime", description="Launch Runtime REPL", show=True),
        Binding(key="v", action="app.create_venv", description="Create VEnv", show=True),
        Binding(key="g", action="app.create_global_venv", description="Create Global VEnv", show=True),
    ]

    if get_managers():
        BINDINGS.extend(
            [
                Binding(key="i", action="app.install_runtime", description="Install New Runtime", show=True),
                Binding(key="delete", action="app.uninstall_runtime", description="Uninstall Runtime", show=True),
            ]
        )

    def __init__(self, *args, config, **kwargs):
        super().__init__(*args, **kwargs)

        self.config = config
        self._runtime_catalogue = {}

    def on_mount(self):
        self.setup_columns()
        self.call_after_refresh(self.load_runtimes, clear_first=False)

    def setup_columns(self):
        self.cursor_type = "row"
        self.add_columns("Version", "Managed By", "Implementation", "Path")

    def runtime_from_key(self, key) -> PythonInstall:
        return self._runtime_catalogue[key]

    @work
    async def load_runtimes(self, clear_first=True):
        self.loading = True
        try:
            if clear_first:
                self.clear()
                self._runtime_catalogue = {}

            loop = asyncio.get_running_loop()
            deduped_installs = await loop.run_in_executor(None, list_installs_deduped)

            for install in deduped_installs:
                self._runtime_catalogue[install.executable] = install

                if install.version_str == install.implementation_version_str:
                    version_str = install.version_str
                else:
                    version_str = f"{install.version_str} / {install.implementation_version_str}"

                managed_by = MANAGED_BY_MAPPING.get(
                    install.managed_by,
                    install.managed_by
                )

                self.add_row(
                    version_str,
                    managed_by,
                    install.implementation,
                    substitute_home(install.executable),
                    key=install.executable
                )
        finally:
            self.refresh_bindings()
            self.loading = False


class ManagerApp(App):
    BINDINGS = [
        Binding(key="q", action="quit", description="Quit"),
    ]

    CSS = """
    .boxed {
        height: auto;
        max-height: 95h;
        border: $primary-darken-2;
    }
    .boxed_fillheight {
        height: 1fr;
        min-height: 40h;
        border: $primary-darken-2;
    }
    .boxed_limitheight {
        height: auto;
        max-height: 60h;
        border: $primary-darken-2;
    }
    .boxed_noborder {
        height: auto;
        border: hidden;
        margin: 1;
    }
    """

    config: Config
    _venv_table: VEnvTable
    _runtime_table: RuntimeTable
    _venv_dependency_cache: dict[str, list[PythonPackage]]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.config = Config.from_file()

        self._venv_table = VEnvTable(config=self.config)
        self._runtime_table = RuntimeTable(config=self.config)
        self._runtime_table.styles.height = "1fr"

        self._venv_dependency_cache = {}

    def on_mount(self):
        self.title = f"Ducktools.PyTUI v{app_version}: Python Environment and Runtime Manager"

    def compose(self):
        yield Header()
        with Vertical(classes="boxed_limitheight"):
            yield Label("Virtual Environments")
            yield self._venv_table
        with Vertical(classes="boxed_fillheight"):
            yield Label("Python Runtimes")
            yield self._runtime_table
        yield Footer()

    @property
    def selected_venv(self) -> PythonVEnv | None:
        table = self._venv_table

        try:
            row = table.coordinate_to_cell_key(table.cursor_coordinate)
        except CellDoesNotExist:
            return None

        return table.venv_from_key(row.row_key.value)

    @property
    def selected_runtime(self) -> PythonInstall | None:
        table = self._runtime_table

        try:
            row = table.coordinate_to_cell_key(table.cursor_coordinate)
        except CellDoesNotExist:
            return None

        return table.runtime_from_key(row.row_key.value)

    def action_launch_runtime(self):
        runtime = self.selected_runtime
        if runtime is None:
            return

        # Suspend the app and launch python
        # Ignore keyboard interrupts otherwise the program will exit when this exits.
        with self.suspend():
            launch_repl(runtime.executable)

        # Redraw
        self.refresh()

    def action_launch_venv_repl(self):
        venv = self.selected_venv
        if venv is None:
            return

        # Suspend the app and launch python
        # Ignore keyboard interrupts otherwise the program will exit when this exits.
        with self.suspend():
            launch_repl(venv.executable)

        # Redraw
        self.refresh()

    @work
    async def action_list_venv_packages(self):
        venv = self.selected_venv
        if venv is None:
            self.notify("No VEnv Selected", severity="warning")
            return
        elif venv.version < (3, 9):
            self.notify(
                f"Package listing not supported for Python {venv.version_str}",
                severity="warning",
            )
            return

        dependency_cache = self._venv_dependency_cache.get(venv.folder)
        dependency_screen = DependencyScreen(venv=venv, dependency_cache=dependency_cache)

        dependencies = await self.push_screen_wait(dependency_screen)
        self._venv_dependency_cache[venv.folder] = dependencies

    def action_activated_shell(self):
        venv = self.selected_venv
        if venv is None:
            self.notify("No VEnv selected")
            return

        if self.config.shell is None:
            self.notify("Failed to find known shell on PATH")
            return

        with self.suspend():
            launch_shell(venv, self.config.shell)

        # Redraw
        self.refresh()

    def _check_runtime_for_venv(self, runtime: PythonInstall | None) -> bool:
        """
        Check if a runtime is selected and if it can be used to create a venv

        :param runtime: PythonInstall object
        :return:
        """
        if runtime is None:
            self.notify("No runtime selected.", severity="warning")
            return False

        if runtime.implementation.lower() == "micropython":
            self.notify(
                "MicroPython does not support VEnv creation.",
                title="Error",
                severity="error",
            )
            return False
        elif runtime.version < (3, 4):
            self.notify(
                f"ducktools-pytui does not support VEnv creation for Python {runtime.version_str}",
                title="Error",
                severity="error",
            )
            return False

        return True

    async def _build_venv(
        self,
        runtime: PythonInstall,
        venv_path: str,
        global_venv: bool
    ) -> None:
        """
        Call python and create the actual venv

        :param runtime:
        :param venv_path:
        :return:
        """
        self._venv_table.loading = True
        loop = asyncio.get_running_loop()
        try:
            new_venv = await loop.run_in_executor(
                None,
                create_venv,
                runtime, venv_path, self.config.include_pip, self.config.latest_pip,
            )
        except FileExistsError:
            self.notify(
                f"Failed to create venv {venv_path!r}, folder already exists",
                title="Error",
                severity="error",
            )
        except subprocess.CalledProcessError as e:
            self.notify(f"Failed to create venv {venv_path!r}. Process Error: {e}")
        else:
            self.notify(f"VEnv {venv_path!r} created", title="Success")
            self._venv_table.add_venv(new_venv, sort=True, global_venv=global_venv)
        finally:
            self._venv_table.loading = False

    @work
    async def action_create_venv(self):
        runtime = self.selected_runtime
        if self._check_runtime_for_venv(runtime) is False:
            return

        venv_screen = VEnvCreateScreen(runtime=runtime)
        venv_name = await self.push_screen_wait(venv_screen)

        if venv_name is None:
            return
        elif venv_name == "":
            venv_name = ".venv"

        await self._build_venv(runtime, venv_name, global_venv=False)

    @work
    async def action_create_global_venv(self):
        runtime = self.selected_runtime
        if self._check_runtime_for_venv(runtime) is False:
            return

        venv_screen = VEnvCreateScreen(runtime=runtime, global_venv=True)
        venv_name = await self.push_screen_wait(venv_screen)

        if venv_name is None:
            return

        venv_path = os.path.join(self.config.global_venv_folder, venv_name)

        self._venv_table.loading = True

        await self._build_venv(runtime, venv_path, global_venv=True)

    def action_delete_venv(self):
        venv = self.selected_venv
        if venv is None:
            return

        if venv.folder == sys.prefix:
            self.notify(
                "Can not delete the VEnv being used to run ducktools-pytui",
                severity="warning",
            )
            return

        delete_venv(venv)
        self._venv_table.remove_venv(venv)
        self._venv_dependency_cache.pop(venv.folder, None)

    @work
    async def action_install_runtime(self):
        if not get_managers():
            return

        runtime_screen = RuntimeInstallScreen()
        runtime = await self.push_screen_wait(runtime_screen)

        if runtime is not None:
            if runtime.will_overwrite:
                # Add a confirmation prompt here
                pass

            self._runtime_table.loading = True
            loop = asyncio.get_running_loop()

            try:
                result = await loop.run_in_executor(None, runtime.install)  # noqa
                # self.notify(result.stderr)
            except FileNotFoundError as e:
                self.notify(
                    f"Install Failed: {markup.escape(str(e))}",
                    title="Manager Not Found",
                    severity="error",
                )
            else:
                if result.returncode == 0:
                    self.notify(
                        f"{runtime.key} installed successfully",
                        title="New Install"
                    )
                    self._runtime_table.load_runtimes(clear_first=True)
                else:
                    for line in result.stderr.split("\n"):
                        if line:
                            self.notify(
                                markup.escape(line.strip()),
                                title="Failed Install",
                                severity="error"
                            )
            finally:
                self._runtime_table.loading = False
                self.set_focus(self._runtime_table)
                self.refresh_bindings()

    @work
    async def action_uninstall_runtime(self):
        if not get_managers():
            return

        runtime = self.selected_runtime
        if runtime is None:
            return

        # Check if the executable is within the base prefix folder
        if runtime.paths.get("stdlib") == sysconfig.get_path("stdlib"):
            self.notify(
                "Can not uninstall the runtime being used to run ducktools-pytui",
                severity="warning",
            )
            return

        listing = find_matching_listing(runtime)
        if listing is None:
            self.notify(
                f"{runtime.executable} is not a managed runtime",
                severity="warning"
            )
            return

        loop = asyncio.get_running_loop()
        self._runtime_table.loading = True
        try:
            result = await loop.run_in_executor(None, listing.uninstall)  # noqa
            # self.notify(result.stderr)
        except FileNotFoundError as e:
            self.notify(
                f"Uninstall Failed: {markup.escape(str(e))}",
                title="Manager Not Found",
                severity="error",
            )
        else:
            if result.returncode == 0:
                self.notify(
                    f"Runtime {listing.key!r} uninstalled."
                )
                self._runtime_table.load_runtimes(clear_first=True)
            else:
                for line in result.stderr.split("\n"):
                    if line:
                        self.notify(
                            markup.escape(line.strip()),
                            title="Failed Uninstall",
                            severity="error"
                        )
        finally:
            self._runtime_table.loading = False
            self.set_focus(self._runtime_table)
            self.refresh_bindings()



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os
import os.path
import shutil
import subprocess

from ducktools.pythonfinder import PythonInstall
from ducktools.pythonfinder.venv import PythonVEnv

from ._version import __version__
from .shells import Shell
from .util import run


WIN_HISTORY_FIXED = False


def fix_win_history():
    """
    Fix the windows history and set a global flag that it has been set
    :return:
    """
    global WIN_HISTORY_FIXED
    from .util.win32_terminal_hist import set_console_history_info
    set_console_history_info()
    WIN_HISTORY_FIXED = True


def launch_repl(python_exe: str) -> None:
    if os.name == "nt" and not WIN_HISTORY_FIXED:
        fix_win_history()
    run([python_exe])  # type: ignore


def create_venv(
    python_runtime: PythonInstall,
    venv_path: str = ".venv",
    include_pip: bool = True,
    latest_pip: bool = True
) -> PythonVEnv:
    # Unlike the regular venv command defaults this will create an environment
    # and download the *newest* pip (assuming the parent venv includes pip)

    if os.path.exists(venv_path):
        raise FileExistsError(f"VEnv '{venv_path}' already exists.")

    python_exe = python_runtime.executable

    # Also always include the pip bundled with graalpy and don't update
    is_graalpy = python_runtime.implementation == "graalpy"

    venv_cmd = [python_exe, "-m", "venv", venv_path]
    if not is_graalpy:
        if not include_pip:
            venv_cmd.append("--without-pip")
        elif latest_pip and python_runtime.version >= (3, 9):
            venv_cmd.append("--upgrade-deps")

    # These tasks run in the background so don't need to block ctrl+c
    # Capture output to not mess with the textual display
    subprocess.run(venv_cmd, capture_output=True, check=True)

    config_path = os.path.join(os.path.realpath(venv_path), "pyvenv.cfg")

    return PythonVEnv.from_cfg(config_path)


def delete_venv(venv: PythonVEnv) -> None:
    shutil.rmtree(venv.folder, ignore_errors=True)


def install_requirements(
    *,
    venv: PythonVEnv,
    requirements_path: str,
    no_deps: bool = False,
) -> None:
    command = [
        venv.executable,
        "-m", "pip",
        "install",
        "-r", requirements_path,
    ]
    if no_deps:
        command.append("--no-deps")

    run(command)  # type: ignore


def launch_shell(venv: PythonVEnv, shell: Shell) -> None:
    # Launch a shell with a virtual environment activated.
    env = os.environ.copy()

    if "dev" in __version__:
        venv_prompt = f"pytui dev: {os.path.basename(venv.folder)}"
    else:
        venv_prompt = f"pytui: {os.path.basename(venv.folder)}"

    venv_bindir = os.path.dirname(venv.executable)

    base_path = shell.get_env_path()
    venv_env_path = shell.get_deduped_path(base_path, venv_bindir)

    # Set the PYTUI versions of PATH/VIRTUAL_ENV/VIRTUAL_ENV_PROMPT
    # These are copied over in the activation script
    # Or included in the env updates
    env["PYTUI_PATH"] = venv_env_path
    env["PYTUI_VIRTUAL_ENV"] = venv.folder
    env["PYTUI_VIRTUAL_ENV_PROMPT"] = venv_prompt

    env.pop("PYTHONHOME", None)

    if os.name == "nt" and not WIN_HISTORY_FIXED:
        fix_win_history()

    cmd, env_updates = shell.get_venv_shell_command(env)

    # Update env with any required environment variables
    env.update(env_updates)

    print("\nVEnv shell from ducktools.pytui: type 'exit' to close")
    run(cmd, env=env)  # type: ignore



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os
import shutil
import sys
from typing import ClassVar, NamedTuple

from ducktools.classbuilder.prefab import Prefab


from .. import _lazy_imports as _laz
from .._version import __version__
from ..platform_paths import SHELL_SCRIPT_FOLDER


class VEnvShellCommand(NamedTuple):
    """
    Class to store the command and environment variables for the venv
    shell creation command.
    """
    cmd: list[str]
    env: dict[str, str]


class Shell(Prefab):
    registry: ClassVar[dict[str, type[Shell]]] = {}

    name: ClassVar[str] = ""  # Pretty Name
    bin_name: ClassVar[str] = ""  # Name of the binary
    exclude: ClassVar[bool] = False

    path: str

    def __init_subclass__(cls):
        if not cls.exclude:
            Shell.registry[cls.bin_name] = cls

    def get_venv_shell_command(
        self,
        env: dict[str, str]
    ) -> VEnvShellCommand:  # pragma: no cover
        raise NotImplementedError("get_venv_shell_command must be implemented in subclasses")

    def get_env_path(self) -> str:
        """
        Get the PATH environment variable

        This exists as a method because some shells will have specific modified PATH
        variables.

        :return: PATH variable string
        """
        return os.environ.get("PATH", "")

    @staticmethod
    def get_deduped_path(path: str, venv_path: str) -> str:
        """
        Get the deduplicated PATH for an activated VENV

        This needs to be a shell dependent method as some shells such as git bash
        will need to have their own management

        :param path: The value of the PATH environment variable
        :param venv_path: The path to the venv's bin or Scripts folder
        :return: A new value for PATH
        """
        deduped_path = []

        # Add venv_path first, so it doesn't end up duplicated
        components = [venv_path, *path.split(os.pathsep)]
        for p in components:
            # Remove a trailing \ or / if it exists
            p = p.rstrip(os.path.sep)
            if p and p not in deduped_path:
                deduped_path.append(p)

        return os.pathsep.join(deduped_path)

    @classmethod
    def which(cls) -> Shell | None:
        """
        Get an instance of this shell type if it is on PATH

        :return: instance of shell type or None if the type is not on PATH
        """
        pth = shutil.which(cls.bin_name)
        if pth:
            return cls(pth)
        return None

    @classmethod
    def from_path(cls, path: str) -> Shell | None:
        name = os.path.basename(path)
        if sys.platform == "win32":
            name = name.lower()  # .EXE can come from which

        shell_type = Shell.registry.get(name)
        if shell_type:
            return shell_type(path)
        return None

    @classmethod
    def get_default(cls) -> Shell | None:
        shell = None
        # Check for SHELL environment variable
        if (shell_env := os.environ.get("SHELL")):
            shell = Shell.from_path(shell_env)

        # Attempt to find shells on PATH
        if shell is None:
            for shell_type in Shell.registry.values():
                shell = shell_type.which()
                if shell:
                    break
            else:
                if sys.platform == "win32":
                    shell_path = os.environ.get("COMSPEC")
                    if shell_path:
                        shell = Shell.from_path(shell_path)

        return shell


def get_shell_script(filename: str) -> str:
    """
    Get the path to the shell script in the SHELL_SCRIPT_FOLDER
    Extract it from the archive or venv folder if needed.

    ALWAYS overwrites the folder if the version does not match even if it is older.

    :param filename: Filename of the script in scripts/
    :return: Path to the file on the system
    """

    shell_script_verfile = os.path.join(SHELL_SCRIPT_FOLDER, ".version")
    valid_verfile = False
    try:
        with open(shell_script_verfile) as f:
            script_ver = f.read()
        if script_ver == __version__:
            valid_verfile = True
    except FileNotFoundError:
        pass

    if not valid_verfile:
        # Delete the old folder
        shutil.rmtree(SHELL_SCRIPT_FOLDER, ignore_errors=True)
        os.makedirs(SHELL_SCRIPT_FOLDER)

        # Get the 'folder' for the scripts directory
        script_folder = _laz.resources.files("ducktools.pytui.shells").joinpath("scripts")
        for script in script_folder.iterdir():
            dest = os.path.join(SHELL_SCRIPT_FOLDER, script.name)
            with _laz.resources.as_file(script) as source:
                shutil.copy(source, dest)

        # Write the version file script
        with open(shell_script_verfile, 'w') as f:
            f.write(__version__)

    script_file = os.path.join(SHELL_SCRIPT_FOLDER, filename)

    if not os.path.exists(script_file):
        raise FileNotFoundError(f"'{filename}' not found in '{SHELL_SCRIPT_FOLDER}'")

    return script_file



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations


from ducktools.pytui.shells import _core as core, cmd


CMD_PATH = R"C:\Windows\System32\cmd.exe"


def test_registered():
    assert cmd.CMDShell.bin_name in core.Shell.registry
    assert cmd.CMDShell is core.Shell.registry[cmd.CMDShell.bin_name]


def test_frompath():
    assert core.Shell.from_path(CMD_PATH) == cmd.CMDShell(CMD_PATH)


def test_get_venv_shell_command():
    venv = R"C:\Users\David\Source\result\.venv"
    path = (
        R"C:\Users\David\Source\result\.venv\Scripts;"
        R"C:\Windows\system32;"
        R"C:\Users\David\.local\bin;"
        R"C:\Program Files\Github CLI"
    )
    prompt = "pytui: .venv"

    env = {
        "PROMPT": "$P$G",
        "PYTUI_PATH": path,
        "PYTUI_VIRTUAL_ENV": venv,
        "PYTUI_VIRTUAL_ENV_PROMPT": prompt,
    }

    cmdshell = cmd.CMDShell(CMD_PATH)

    shell_cmd, env_updates = cmdshell.get_venv_shell_command(env)

    assert shell_cmd == [CMD_PATH, "/k"]
    assert env_updates == {
        "PATH": path,
        "VIRTUAL_ENV": venv,
        "VIRTUAL_ENV_PROMPT": prompt,
        "PROMPT": "(pytui: .venv) $P$G",
    }

    # Test with an old venv already in prompt
    env = {
        "PROMPT": "(pytui: .oldvenv) $P$G",
        "VIRTUAL_ENV_PROMPT": "pytui: .oldvenv",
        "PYTUI_PATH": path,
        "PYTUI_VIRTUAL_ENV": venv,
        "PYTUI_VIRTUAL_ENV_PROMPT": prompt,
    }

    shell_cmd, env_updates = cmdshell.get_venv_shell_command(env)

    assert env_updates == {
        "PATH": path,
        "VIRTUAL_ENV": venv,
        "VIRTUAL_ENV_PROMPT": prompt,
        "PROMPT": "(pytui: .venv) $P$G",
    }


# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# mypy: disable-error-code="attr-defined"
# attributes don't exist on non-windows platform
# but this entire file is only imported on windows.
"""
Helper code to 'fix' the windows terminal history in order to get command history
in Python that doesn't use PyREPL.

Copied from:
https://discuss.python.org/t/interactive-command-history-in-session-started-with-subprocess-on-windows/3701/5
"""
from __future__ import annotations

import ctypes
import collections
from ctypes import wintypes

kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)

HISTORY_NO_DUP_FLAG = 1


class ConsoleHistoryInfoStruct(ctypes.Structure):
    _fields_ = (
        ('cbSize', wintypes.UINT),
        ('HistoryBufferSize', wintypes.UINT),
        ('NumberOfHistoryBuffers', wintypes.UINT),
        ('dwFlags', wintypes.DWORD)
    )

    def __init__(self, *args, **kwds):
        super().__init__(ctypes.sizeof(self), *args, **kwds)


ConsoleHistoryInfo = collections.namedtuple(
    'ConsoleHistoryInfo',
    'bufsize nbuf flags'
)


def get_console_history_info():
    info = ConsoleHistoryInfoStruct()
    if not kernel32.GetConsoleHistoryInfo(ctypes.byref(info)):
        raise ctypes.WinError(ctypes.get_last_error())

    return ConsoleHistoryInfo(
        info.HistoryBufferSize,
        info.NumberOfHistoryBuffers,
        info.dwFlags
    )


def set_console_history_info(
    bufsize=512,
    nbuf=32,
    flags=HISTORY_NO_DUP_FLAG
):
    info = ConsoleHistoryInfoStruct(bufsize, nbuf, flags)
    if not kernel32.SetConsoleHistoryInfo(ctypes.byref(info)):
        raise ctypes.WinError(ctypes.get_last_error())



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from ._core import Shell, VEnvShellCommand


class CMDShell(Shell):
    name = "Command Prompt"
    bin_name = "cmd.exe"

    def get_venv_shell_command(self, env: dict[str, str]) -> VEnvShellCommand:
        shell_prompt = env.get("PROMPT", "$P$G")
        old_venv_prompt = env.get("VIRTUAL_ENV_PROMPT")
        new_venv_prompt = env["PYTUI_VIRTUAL_ENV_PROMPT"]


        if old_venv_prompt and old_venv_prompt in shell_prompt:
            new_prompt = shell_prompt.replace(old_venv_prompt, new_venv_prompt)
        else:
            new_prompt = f"({new_venv_prompt}) {shell_prompt}"

        cmd = [self.path, "/k"]
        env_updates = {
            "PATH": env["PYTUI_PATH"],
            "VIRTUAL_ENV": env["PYTUI_VIRTUAL_ENV"],
            "VIRTUAL_ENV_PROMPT": env["PYTUI_VIRTUAL_ENV_PROMPT"],
            "PROMPT": new_prompt,
        }

        return VEnvShellCommand(cmd, env_updates)



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import functools
import operator
import os.path
import subprocess
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any, ClassVar, Final, Generic, TypeAlias, TypeVar

from ducktools.pythonfinder.shared import PythonInstall, version_str_to_tuple
from ducktools.classbuilder.prefab import prefab, attribute


_version_tuple_type: TypeAlias = tuple[int, int, int, str, int]  # type statement needs 3.12+


@prefab(kw_only=True)
class PythonListing(ABC):
    manager: RuntimeManager

    key: str
    version: str
    implementation: str
    variant: str
    arch: str
    path: str | None
    url: str | None

    # `attribute` is a field specifier as defined in dataclass_transform
    # Not sure why it's not being picked up
    _version_tuple: _version_tuple_type | None = attribute(default=None, private=True)

    @property
    def version_tuple(self) -> _version_tuple_type:
        if not self._version_tuple:
            self._version_tuple = version_str_to_tuple(self.version)
        return self._version_tuple

    @property
    def full_key(self) -> str:
        return f"{type(self).__name__} / {self.key}"

    @property
    def will_overwrite(self) -> bool:
        return False

    @classmethod
    @abstractmethod
    def from_dict(cls, manager, entry) -> PythonListing:
        ...

    @abstractmethod
    def install(self) -> subprocess.CompletedProcess | None:
        ...

    @abstractmethod
    def uninstall(self) -> subprocess.CompletedProcess | None:
        ...


Listing = TypeVar("Listing", bound=PythonListing)


class RuntimeManager(Generic[Listing], ABC):
    available_managers: Final[list[type[RuntimeManager]]] = []
    
    organisation: ClassVar[str]

    def __init_subclass__(cls) -> None:
        RuntimeManager.available_managers.append(cls)

    @staticmethod
    def sort_listings(listings: Iterable[Listing]) -> list[Listing]:
        new_listings = sorted(listings, key=operator.attrgetter("variant", "arch", "key"))
        new_listings.sort(key=operator.attrgetter("version_tuple"), reverse=True)
        new_listings.sort(key=operator.attrgetter("implementation"))
        return new_listings

    @functools.cached_property
    @abstractmethod
    def executable(self) -> str | None:
        """
        Get the path to the manager executable or None if it is not installed
        """
        ...

    @abstractmethod
    def fetch_installed(self) -> list[Listing]:
        """
        Get a list of installed runtimes managed by the manager
        """
        ...

    @abstractmethod
    def _get_download_cache(self) -> list[Listing]:
        """
        List all available downloads (cached method)
        """

    @abstractmethod
    def fetch_downloads(self) -> list[Listing]:
        """
        List available downloads, exclude already downloaded (not cached)
        """

    def find_matching_listing(self, install: PythonInstall) -> Listing | None:
        if install.managed_by is None or not install.managed_by.startswith(self.organisation):
            return None

        # Executable names may not match, one may find python.exe, the other pypy.exe
        # Use the parent folder.
        installed_dict = {
            os.path.dirname(os.path.abspath(py.path)): py
            for py in self.fetch_installed()
            if py.path is not None
        }

        install_path = os.path.dirname(install.executable)

        return installed_dict.get(install_path, None)



import json
from pathlib import Path
from unittest.mock import patch

from pytest import fixture

from ducktools.pythonfinder import PythonInstall
from ducktools.pythonfinder.venv import PythonVEnv

from ducktools.pytui.config import Config


DATA_FOLDER = Path(__file__).parents[1] / "example_data" / "pythonfinder"


@fixture
async def patched_config():
    with (
        patch.object(Config, "from_file") as patched_config,
        patch.object(Config, "write_config"),  # prevent accidental writes
    ):
        # This file shouldn't actually be created as the write is patched
        config = Config(config_file="/tmp/pytui_test_config.json")
        patched_config.return_value = config
        yield config


@fixture(scope="session")
async def runtimes() -> list[PythonInstall]:
    raw_installs = json.loads((DATA_FOLDER / "runtimes_data.json").read_text())
    pys = [PythonInstall.from_json(**inst) for inst in raw_installs]
    assert len(pys) > 0
    return pys


@fixture(scope="session")
async def local_venvs() -> list[PythonVEnv]:
    raw_venvs = json.loads((DATA_FOLDER / "local_venvs.json").read_text())
    venvs = []
    for v in raw_venvs:
        venv = PythonVEnv(
            folder=v["folder"],
            executable=v["executable"],
            version=tuple(v["version"]),
            parent_path=v["parent_path"],
            _parent_executable=v["_parent_executable"],
        )
        venvs.append(venv)
    return venvs


@fixture(scope="session")
async def global_venvs() -> list[PythonVEnv]:
    raw_venvs = json.loads((DATA_FOLDER / "global_venvs.json").read_text())
    venvs = []
    for v in raw_venvs:
        venv = PythonVEnv(
            folder=v["folder"],
            executable=v["executable"],
            version=tuple(v["version"]),
            parent_path=v["parent_path"],
            _parent_executable=v["_parent_executable"],
        )
        venvs.append(venv)
    return venvs


@fixture(autouse=True)
async def patch_list_installs(runtimes):
    with patch("ducktools.pytui.ui.list_installs_deduped") as mock_deduped:
        mock_deduped.return_value = runtimes
        yield mock_deduped


@fixture(autouse=True)
async def patch_list_venvs(local_venvs, global_venvs, patched_config):
    def get_venv(base_dir=None, recursive=False, search_parent_folders=False):
        if base_dir == patched_config.global_venv_folder:
            return global_venvs
        else:
            return local_venvs


    with patch("ducktools.pytui.ui.list_python_venvs") as venv_mock:
        venv_mock.side_effect = get_venv
        yield venv_mock



import sys
from operator import itemgetter
from unittest.mock import patch

import pytest

import ducktools.pytui.ui as ui
from ducktools.pytui.ui import ManagerApp, MANAGED_BY_MAPPING, substitute_home

from textual.worker import WorkerFailed


@patch("ducktools.pytui.ui.HOME", "/home/ducksual")
@pytest.mark.flaky(reruns=5)
async def test_runtime_table(runtimes, patch_list_installs):
    app = ManagerApp()
    async with app.run_test() as pilot:
        await pilot.pause()  # Wait for processing

        patch_list_installs.assert_called()

        assert app._runtime_table.row_count > 0
        row_data = [app._runtime_table.get_row_at(i) for i in range(0, app._runtime_table.row_count)]
        row_data.sort(key=itemgetter(3))  # Sort by Path

        expected = [
            [
                py.version_str if py.implementation_version_str == py.version_str else f"{py.version_str} / {py.implementation_version_str}",
                MANAGED_BY_MAPPING.get(py.managed_by, py.managed_by),
                py.implementation,
                substitute_home(py.executable)
            ]
            for py in runtimes
        ]
        expected.sort(key=itemgetter(3))
        assert row_data == expected



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import os
import subprocess
import sys

from ._core import Shell, VEnvShellCommand, get_shell_script


class BashShell(Shell):
    """
    Shell code for the Linux bash shell
    """
    name = "Bash"
    bin_name = "bash"
    exclude = (sys.platform == "win32")

    def get_venv_shell_command(self, env: dict[str, str]) -> VEnvShellCommand:
        rcfile = get_shell_script("activate_pytui.sh")
        cmd = [self.path, "--rcfile", rcfile]
        env_updates: dict[str, str] = {}
        return VEnvShellCommand(cmd, env_updates)


class GitBashShell(BashShell):
    """
    Shell code for a Git Bash shell on Windows
    """
    name = "Git Bash"
    bin_name = "bash.exe"
    exclude = (sys.platform != "win32")

    def get_env_path(self) -> str:
        """
        Get the git bash environment Path

        This requires launching the shell and echoing
        the default $PATH variable

        :return: Git bash formatted PATH variable
        """
        prompt_getter = subprocess.run(
            [self.path, "-ic", "echo $PATH"],
            text=True,
            capture_output=True
        )
        return prompt_getter.stdout.strip()

    @staticmethod
    def get_deduped_path(path, venv_dir: str) -> str:
        """
        Special dedupe handling for git bash PATH details

        :param path: git bash $PATH output
        :param venv_dir: regular windows style path to VENV
        :return: new git bash style path without duplicates
        """
        # The PATH environment variable on git bash needs to look more like a linux path
        drive, venv_dir = os.path.splitdrive(venv_dir)
        if ":" in drive:
            drive = drive.replace(":", "").lower()
            drive = f"/{drive}"
        venv_dir = venv_dir.replace("\\", "/")
        new_venv_bindir = "".join([drive, venv_dir])

        deduped_path = []

        components = [new_venv_bindir, *path.split(":")]
        for p in components:
            p = p.rstrip("/")  # Remove trailing slash
            if p and p not in deduped_path:
                deduped_path.append(p)

        return ":".join(deduped_path)



import sys

import pytest

from ducktools.pytui.ui import substitute_home


@pytest.mark.parametrize(
        "path, expected", 
        [
            ("/usr/bin/python", None),  # None is used for unchanged
            ("/home/david/.local/share/ducktools", "~/.local/share/ducktools"),
            ("/home/other/.local/share/ducktools", None),
        ],
    )
def test_substitute_home(path, expected):
    homedir = "/home/david"

    if expected is None or sys.platform == "win32":
        expected = path

    assert substitute_home(path, homedir=homedir) == expected



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Here we can use pathlib as it's a build script
import sys

import shutil
import subprocess
import zipapp

from pathlib import Path


project_path = Path(__file__).parents[1]
project_build_path = project_path / "build" / "project_build"
build_path = project_path / "build" / "application"
build_bin = build_path / "bin"
dist_path = project_path / "dist" / "pytui.pyz"
main_func = "ducktools.pytui.__main__:main"


def main():
    # Prepare install and build paths
    shutil.rmtree(build_path, ignore_errors=True)
    shutil.rmtree(project_build_path, ignore_errors=True)
    dist_path.unlink(missing_ok=True)

    build_path.mkdir(parents=True)
    dist_path.parent.mkdir(parents=True, exist_ok=True)

    # Download locked dependencies
    subprocess.run(
        [
            sys.executable, "-m",
            "pip", "install",
            "--no-compile",
            "-r", str(project_path / "requirements.txt"),
            "--target", str(build_path),
        ]
    )

    # Install just the current package in another folder
    subprocess.run(
        [
            sys.executable, "-m",
            "pip", "install",
            "--no-compile", str(project_path),
            "--no-deps",
            "--target", str(project_build_path)
        ]
    )

    shutil.move(
        project_build_path / "ducktools" / "pytui",
        build_path / "ducktools" / "pytui",
    )

    for fld in project_build_path.glob("ducktools_pytui*/"):
        shutil.move(str(fld), str(build_path))  # convert to str for 3.8

    # Remove script files
    shutil.rmtree(build_bin)
    shutil.rmtree(project_build_path)

    zipapp.create_archive(
        build_path,
        dist_path,
        interpreter="/usr/bin/env python3",
        main=main_func,
    )


if __name__ == "__main__":
    main()



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Handle the Windows python PyManager
from __future__ import annotations

import functools
import json
import os.path
import platform
import re
import shutil
import subprocess

from typing import Any, ClassVar

from ducktools.classbuilder.prefab import prefab

from .base import RuntimeManager, PythonListing


freethreaded_re = re.compile(r"^\d+.\d+t.*$")


@prefab(kw_only=True)
class PythonCoreListing(PythonListing):
    manager: PythonCoreManager

    name: str
    tag: str
    company: str
    install_for: list[str]

    @classmethod
    def from_dict(cls, manager: PythonCoreManager, entry: dict[str, Any]) -> PythonCoreListing:
    
        key = entry["id"]
        version = entry["sort-version"]
        name = entry["display-name"]
        company = entry["company"]

        implementation = "cpython"
        tag = entry["tag"]
        install_for = entry["install-for"]
        variant = "freethreaded" if freethreaded_re.match(tag) else "default"
        path = entry.get("executable")
        url = entry.get("url")

        # Fix up non-existent paths
        if not (path and os.path.exists(path)):
            path = None

        if "-32" in key:
            arch = "x86"
        elif "-arm64" in key:
            arch = "ARM64"
        else:
            arch = "x86_64"

        return cls(
            manager=manager,
            key=key,
            version=version,
            name=name,
            company=company,
            implementation=implementation,
            variant=variant,
            tag=tag,
            install_for=install_for,
            path=path,
            url=url,
            arch=arch,
        )

    @property
    def will_overwrite(self) -> bool:
        for v in self.manager.fetch_installed():
            if self.tag == v.tag:
                return True
        return False

    def install(self) -> subprocess.CompletedProcess | None:
        if self.manager.executable is None:
            raise FileNotFoundError("Could not find the 'pymanager' executable on PATH")

        if self.path:
            return None

        # Guard against empty list
        if self.install_for:
            tag = self.install_for[0]
        else:
            tag = self.tag

        cmd = [
            self.manager.executable, "install", tag, "-y",
        ]

        result = subprocess.run(
            cmd, capture_output=True, text=True
        )
        return result


    def uninstall(self) -> subprocess.CompletedProcess | None:
        if self.manager.executable is None:
            raise FileNotFoundError("Could not find the 'pymanager' executable on PATH")

        if not (self.path and os.path.exists(self.path)):
            return None

        # Guard against empty list
        if self.install_for:
            tag = self.install_for[0]
        else:
            tag = self.tag

        cmd = [
            self.manager.executable, "uninstall", tag, "-y",
        ]
        result = subprocess.run(
            cmd, capture_output=True, text=True
        )
        return result


class PythonCoreManager(RuntimeManager[PythonCoreListing]):
    organisation: ClassVar[str] = "PythonCore"

    @functools.cached_property
    def executable(self) -> str | None:
        return shutil.which("pymanager")

    def fetch_installed(self) -> list[PythonCoreListing]:
        if self.executable is None:
            raise FileNotFoundError("Could not find the 'pymanager' executable on PATH")
        
        cmd = [
            self.executable, "list", "--only-managed", "--format=json",
        ]
        installed_list_cmd = subprocess.run(
            cmd, capture_output=True, text=True, check=True,
        )
        json_data = json.loads(installed_list_cmd.stdout)
        installed_pys = [
            PythonCoreListing.from_dict(manager=self, entry=v)
            for v in json_data.get("versions", [])
        ]
        return installed_pys

    @functools.lru_cache(maxsize=None)
    def _get_download_cache(self) -> list[PythonCoreListing]:
        """
        Get the raw unfiltered download list.
        This is cached to avoid repeated calls to PyManager for downloads.
        This data should only change on new Python releases.

        :return: List of PythonCoreListings for all downloads
        """
        if self.executable is None:
            raise FileNotFoundError("Could not find the 'pymanager' executable on PATH")
        
        cmd = [
            self.executable, "list", "--online", "--format=json",
        ]
        download_list_cmd = subprocess.run(
            cmd, capture_output=True, text=True, check=True,
        )
        json_data = json.loads(download_list_cmd.stdout)

        downloads = [
            PythonCoreListing.from_dict(manager=self, entry=v)
            for v in json_data.get("versions", [])
        ]

        return downloads

    def fetch_downloads(self) -> list[PythonCoreListing]:
        """
        Get the filtered list of downloads, with installed versions removed.

        :return: filtered download list
        """
        downloads = self._get_download_cache()

        installed_versions = {(v.key, v.version) for v in self.fetch_installed()}

        # PythonEmbed used for embedded distributions
        # PythonTest used for distributions with tests
        # PythonCore are the ones we want
        arch = platform.machine()
        if arch == "AMD64":
            arch = "x86_64"

        download_listings = [
            v
            for v in downloads
            if (v.key, v.version) not in installed_versions
            and v.company == "PythonCore"
            and v.arch == arch
        ]

        return self.sort_listings(download_listings)
